"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.OpenId4VcIssuerService = void 0;
const core_1 = require("@credo-ts/core");
const oid4vci_common_1 = require("@sphereon/oid4vci-common");
const oid4vci_issuer_1 = require("@sphereon/oid4vci-issuer");
const shared_1 = require("../shared");
const router_1 = require("../shared/router");
const transform_1 = require("../shared/transform");
const utils_1 = require("../shared/utils");
const OpenId4VcIssuerModuleConfig_1 = require("./OpenId4VcIssuerModuleConfig");
const repository_1 = require("./repository");
const w3cOpenId4VcFormats = [
    shared_1.OpenId4VciCredentialFormatProfile.JwtVcJson,
    shared_1.OpenId4VciCredentialFormatProfile.JwtVcJsonLd,
    shared_1.OpenId4VciCredentialFormatProfile.LdpVc,
];
/**
 * @internal
 */
let OpenId4VcIssuerService = class OpenId4VcIssuerService {
    constructor(w3cCredentialService, jwsService, openId4VcIssuerConfig, openId4VcIssuerRepository) {
        this.getJwtVerifyCallback = (agentContext) => {
            return async (opts) => {
                let didDocument = undefined;
                const { isValid, jws } = await this.jwsService.verifyJws(agentContext, {
                    jws: opts.jwt,
                    // Only handles kid as did resolution. JWK is handled by jws service
                    jwkResolver: async ({ protectedHeader: { kid } }) => {
                        if (!kid)
                            throw new core_1.CredoError('Missing kid in protected header.');
                        if (!kid.startsWith('did:'))
                            throw new core_1.CredoError('Only did is supported for kid identifier');
                        const didsApi = agentContext.dependencyManager.resolve(core_1.DidsApi);
                        didDocument = await didsApi.resolveDidDocument(kid);
                        const verificationMethod = didDocument.dereferenceKey(kid, ['authentication', 'assertionMethod']);
                        const key = (0, core_1.getKeyFromVerificationMethod)(verificationMethod);
                        return (0, core_1.getJwkFromKey)(key);
                    },
                });
                if (!isValid)
                    throw new core_1.CredoError('Could not verify JWT signature.');
                // TODO: the jws service should return some better decoded metadata also from the resolver
                // as currently is less useful if you afterwards need properties from the JWS
                const firstJws = jws.signatures[0];
                const protectedHeader = core_1.JsonEncoder.fromBase64(firstJws.protected);
                return {
                    jwt: { header: protectedHeader, payload: core_1.JsonEncoder.fromBase64(jws.payload) },
                    kid: protectedHeader.kid,
                    jwk: protectedHeader.jwk ? (0, core_1.getJwkFromJson)(protectedHeader.jwk) : undefined,
                    did: didDocument === null || didDocument === void 0 ? void 0 : didDocument.id,
                    alg: protectedHeader.alg,
                    didDocument,
                };
            };
        };
        this.getSdJwtVcCredentialSigningCallback = (agentContext, options) => {
            return async () => {
                const sdJwtVcApi = agentContext.dependencyManager.resolve(core_1.SdJwtVcApi);
                const sdJwtVc = await sdJwtVcApi.sign(options);
                return (0, transform_1.getSphereonVerifiableCredential)(sdJwtVc);
            };
        };
        this.getW3cCredentialSigningCallback = (agentContext, options) => {
            return async (opts) => {
                const { jwtVerifyResult, format } = opts;
                const { kid, didDocument: holderDidDocument } = jwtVerifyResult;
                if (!kid)
                    throw new core_1.CredoError('Missing Kid. Cannot create the holder binding');
                if (!holderDidDocument)
                    throw new core_1.CredoError('Missing did document. Cannot create the holder binding.');
                if (!format)
                    throw new core_1.CredoError('Missing format. Cannot issue credential.');
                const formatMap = {
                    [shared_1.OpenId4VciCredentialFormatProfile.JwtVcJson]: core_1.ClaimFormat.JwtVc,
                    [shared_1.OpenId4VciCredentialFormatProfile.JwtVcJsonLd]: core_1.ClaimFormat.JwtVc,
                    [shared_1.OpenId4VciCredentialFormatProfile.LdpVc]: core_1.ClaimFormat.LdpVc,
                };
                const w3cServiceFormat = formatMap[format];
                // Set the binding on the first credential subject if not set yet
                // on any subject
                if (!options.credential.credentialSubjectIds.includes(holderDidDocument.id)) {
                    const credentialSubject = Array.isArray(options.credential.credentialSubject)
                        ? options.credential.credentialSubject[0]
                        : options.credential.credentialSubject;
                    credentialSubject.id = holderDidDocument.id;
                }
                const didsApi = agentContext.dependencyManager.resolve(core_1.DidsApi);
                const issuerDidDocument = await didsApi.resolveDidDocument(options.verificationMethod);
                const verificationMethod = issuerDidDocument.dereferenceVerificationMethod(options.verificationMethod);
                if (w3cServiceFormat === core_1.ClaimFormat.JwtVc) {
                    const key = (0, core_1.getKeyFromVerificationMethod)(verificationMethod);
                    const alg = (0, core_1.getJwkFromKey)(key).supportedSignatureAlgorithms[0];
                    if (!alg) {
                        throw new core_1.CredoError(`No supported JWA signature algorithms for key type ${key.keyType}`);
                    }
                    const signed = await this.w3cCredentialService.signCredential(agentContext, {
                        format: w3cServiceFormat,
                        credential: options.credential,
                        verificationMethod: options.verificationMethod,
                        alg,
                    });
                    return (0, transform_1.getSphereonVerifiableCredential)(signed);
                }
                else {
                    const key = (0, core_1.getKeyFromVerificationMethod)(verificationMethod);
                    const proofType = (0, utils_1.getProofTypeFromKey)(agentContext, key);
                    const signed = await this.w3cCredentialService.signCredential(agentContext, {
                        format: w3cServiceFormat,
                        credential: options.credential,
                        verificationMethod: options.verificationMethod,
                        proofType: proofType,
                    });
                    return (0, transform_1.getSphereonVerifiableCredential)(signed);
                }
            };
        };
        this.getCredentialDataSupplier = (agentContext, options) => {
            return async (args) => {
                var _a;
                const { credentialRequest, credentialOffer } = args;
                const issuerMetadata = this.getIssuerMetadata(agentContext, options.issuer);
                const offeredCredentialsMatchingRequest = this.findOfferedCredentialsMatchingRequest(credentialOffer.credential_offer, credentialRequest, issuerMetadata.credentialsSupported);
                if (offeredCredentialsMatchingRequest.length === 0) {
                    throw new core_1.CredoError('No offered credentials match the credential request.');
                }
                if (offeredCredentialsMatchingRequest.length > 1) {
                    agentContext.config.logger.debug('Multiple credentials from credentials supported matching request, picking first one.');
                }
                const holderBinding = await this.getHolderBindingFromRequest(credentialRequest);
                const mapper = (_a = options.credentialRequestToCredentialMapper) !== null && _a !== void 0 ? _a : this.openId4VcIssuerConfig.credentialEndpoint.credentialRequestToCredentialMapper;
                const signOptions = await mapper({
                    agentContext,
                    holderBinding,
                    credentialOffer,
                    credentialRequest,
                    credentialsSupported: offeredCredentialsMatchingRequest,
                });
                if (signOptions.format === core_1.ClaimFormat.JwtVc || signOptions.format === core_1.ClaimFormat.LdpVc) {
                    if (!w3cOpenId4VcFormats.includes(credentialRequest.format)) {
                        throw new core_1.CredoError(`The credential to be issued does not match the request. Cannot issue a W3cCredential if the client expects a credential of format '${credentialRequest.format}'.`);
                    }
                    return {
                        format: credentialRequest.format,
                        credential: core_1.JsonTransformer.toJSON(signOptions.credential),
                        signCallback: this.getW3cCredentialSigningCallback(agentContext, signOptions),
                    };
                }
                else if (signOptions.format === core_1.ClaimFormat.SdJwtVc) {
                    if (credentialRequest.format !== shared_1.OpenId4VciCredentialFormatProfile.SdJwtVc) {
                        throw new core_1.CredoError(`Invalid credential format. Expected '${shared_1.OpenId4VciCredentialFormatProfile.SdJwtVc}', received '${credentialRequest.format}'.`);
                    }
                    if (credentialRequest.vct !== signOptions.payload.vct) {
                        throw new core_1.CredoError(`The types of the offered credentials do not match the types of the requested credential. Offered '${signOptions.payload.vct}' Requested '${credentialRequest.vct}'.`);
                    }
                    return {
                        format: credentialRequest.format,
                        // NOTE: we don't use the credential value here as we pass the credential directly to the singer
                        credential: Object.assign({}, signOptions.payload),
                        signCallback: this.getSdJwtVcCredentialSigningCallback(agentContext, signOptions),
                    };
                }
                else {
                    throw new core_1.CredoError(`Unsupported credential format`);
                }
            };
        };
        this.w3cCredentialService = w3cCredentialService;
        this.jwsService = jwsService;
        this.openId4VcIssuerConfig = openId4VcIssuerConfig;
        this.openId4VcIssuerRepository = openId4VcIssuerRepository;
    }
    getIssuerMetadata(agentContext, issuerRecord) {
        const config = agentContext.dependencyManager.resolve(OpenId4VcIssuerModuleConfig_1.OpenId4VcIssuerModuleConfig);
        const issuerUrl = (0, core_1.joinUriParts)(config.baseUrl, [issuerRecord.issuerId]);
        const issuerMetadata = {
            issuerUrl,
            tokenEndpoint: (0, core_1.joinUriParts)(issuerUrl, [config.accessTokenEndpoint.endpointPath]),
            credentialEndpoint: (0, core_1.joinUriParts)(issuerUrl, [config.credentialEndpoint.endpointPath]),
            credentialsSupported: issuerRecord.credentialsSupported,
            issuerDisplay: issuerRecord.display,
        };
        return issuerMetadata;
    }
    async createCredentialOffer(agentContext, options) {
        const { preAuthorizedCodeFlowConfig, authorizationCodeFlowConfig, issuer, offeredCredentials } = options;
        const vcIssuer = this.getVcIssuer(agentContext, issuer);
        // this checks if the structure of the credentials is correct
        // it throws an error if a offered credential cannot be found in the credentialsSupported
        (0, shared_1.getOfferedCredentials)(options.offeredCredentials, vcIssuer.issuerMetadata.credentials_supported);
        const { uri, session } = await vcIssuer.createCredentialOfferURI({
            grants: await this.getGrantsFromConfig(agentContext, preAuthorizedCodeFlowConfig, authorizationCodeFlowConfig),
            credentials: offeredCredentials,
            // TODO: support hosting of credential offers within AFJ
            credentialOfferUri: options.hostedCredentialOfferUrl,
            baseUri: options.baseUri,
        });
        const credentialOfferPayload = session.credentialOffer.credential_offer;
        return {
            credentialOfferPayload,
            credentialOffer: uri,
        };
    }
    async getCredentialOfferFromUri(agentContext, uri) {
        const { credentialOfferSessionId, credentialOfferSession } = await this.getCredentialOfferSessionFromUri(agentContext, uri);
        credentialOfferSession.lastUpdatedAt = +new Date();
        credentialOfferSession.status = oid4vci_common_1.IssueStatus.OFFER_URI_RETRIEVED;
        await this.openId4VcIssuerConfig
            .getCredentialOfferSessionStateManager(agentContext)
            .set(credentialOfferSessionId, credentialOfferSession);
        return credentialOfferSession.credentialOffer.credential_offer;
    }
    async createCredentialResponse(agentContext, options) {
        const { credentialRequest, issuer } = options;
        if (!credentialRequest.proof)
            throw new core_1.CredoError('No proof defined in the credentialRequest.');
        const vcIssuer = this.getVcIssuer(agentContext, issuer);
        const issueCredentialResponse = await vcIssuer.issueCredential({
            credentialRequest,
            tokenExpiresIn: this.openId4VcIssuerConfig.accessTokenEndpoint.tokenExpiresInSeconds,
            // This can just be combined with signing callback right?
            credentialDataSupplier: this.getCredentialDataSupplier(agentContext, options),
            newCNonce: undefined,
            responseCNonce: undefined,
        });
        if (!issueCredentialResponse.credential) {
            throw new core_1.CredoError('No credential found in the issueCredentialResponse.');
        }
        if (issueCredentialResponse.acceptance_token) {
            throw new core_1.CredoError('Acceptance token not yet supported.');
        }
        return issueCredentialResponse;
    }
    async getAllIssuers(agentContext) {
        return this.openId4VcIssuerRepository.getAll(agentContext);
    }
    async getByIssuerId(agentContext, issuerId) {
        return this.openId4VcIssuerRepository.getByIssuerId(agentContext, issuerId);
    }
    async updateIssuer(agentContext, issuer) {
        return this.openId4VcIssuerRepository.update(agentContext, issuer);
    }
    async createIssuer(agentContext, options) {
        var _a;
        // TODO: ideally we can store additional data with a key, such as:
        // - createdAt
        // - purpose
        const accessTokenSignerKey = await agentContext.wallet.createKey({
            keyType: core_1.KeyType.Ed25519,
        });
        const openId4VcIssuer = new repository_1.OpenId4VcIssuerRecord({
            issuerId: (_a = options.issuerId) !== null && _a !== void 0 ? _a : core_1.utils.uuid(),
            display: options.display,
            accessTokenPublicKeyFingerprint: accessTokenSignerKey.fingerprint,
            credentialsSupported: options.credentialsSupported,
        });
        await this.openId4VcIssuerRepository.save(agentContext, openId4VcIssuer);
        await (0, router_1.storeActorIdForContextCorrelationId)(agentContext, openId4VcIssuer.issuerId);
        return openId4VcIssuer;
    }
    async rotateAccessTokenSigningKey(agentContext, issuer) {
        const accessTokenSignerKey = await agentContext.wallet.createKey({
            keyType: core_1.KeyType.Ed25519,
        });
        // TODO: ideally we can remove the previous key
        issuer.accessTokenPublicKeyFingerprint = accessTokenSignerKey.fingerprint;
        await this.openId4VcIssuerRepository.update(agentContext, issuer);
    }
    async getCredentialOfferSessionFromUri(agentContext, uri) {
        var _a;
        const uriState = await this.openId4VcIssuerConfig.getUriStateManager(agentContext).get(uri);
        if (!uriState)
            throw new core_1.CredoError(`Credential offer uri '${uri}' not found.`);
        const credentialOfferSessionId = (_a = uriState.preAuthorizedCode) !== null && _a !== void 0 ? _a : uriState.issuerState;
        if (!credentialOfferSessionId) {
            throw new core_1.CredoError(`Credential offer uri '${uri}' is not associated with a preAuthorizedCode or issuerState.`);
        }
        const credentialOfferSession = await this.openId4VcIssuerConfig
            .getCredentialOfferSessionStateManager(agentContext)
            .get(credentialOfferSessionId);
        if (!credentialOfferSession)
            throw new core_1.CredoError(`Credential offer session for '${uri}' with id '${credentialOfferSessionId}' not found.`);
        return { credentialOfferSessionId, credentialOfferSession };
    }
    getVcIssuer(agentContext, issuer) {
        const issuerMetadata = this.getIssuerMetadata(agentContext, issuer);
        const builder = new oid4vci_issuer_1.VcIssuerBuilder()
            .withCredentialIssuer(issuerMetadata.issuerUrl)
            .withCredentialEndpoint(issuerMetadata.credentialEndpoint)
            .withTokenEndpoint(issuerMetadata.tokenEndpoint)
            .withCredentialsSupported(issuerMetadata.credentialsSupported)
            .withCNonceStateManager(this.openId4VcIssuerConfig.getCNonceStateManager(agentContext))
            .withCredentialOfferStateManager(this.openId4VcIssuerConfig.getCredentialOfferSessionStateManager(agentContext))
            .withCredentialOfferURIStateManager(this.openId4VcIssuerConfig.getUriStateManager(agentContext))
            .withJWTVerifyCallback(this.getJwtVerifyCallback(agentContext))
            .withCredentialSignerCallback(() => {
            throw new core_1.CredoError('Credential signer callback should be overwritten. This is a no-op');
        });
        if (issuerMetadata.authorizationServer) {
            builder.withAuthorizationServer(issuerMetadata.authorizationServer);
        }
        if (issuerMetadata.issuerDisplay) {
            builder.withIssuerDisplay(issuerMetadata.issuerDisplay);
        }
        return builder.build();
    }
    async getGrantsFromConfig(agentContext, preAuthorizedCodeFlowConfig, authorizationCodeFlowConfig) {
        var _a, _b, _c;
        if (!preAuthorizedCodeFlowConfig && !authorizationCodeFlowConfig) {
            throw new core_1.CredoError(`Either preAuthorizedCodeFlowConfig or authorizationCodeFlowConfig must be provided.`);
        }
        const grants = {
            'urn:ietf:params:oauth:grant-type:pre-authorized_code': preAuthorizedCodeFlowConfig && {
                'pre-authorized_code': (_a = preAuthorizedCodeFlowConfig.preAuthorizedCode) !== null && _a !== void 0 ? _a : (await agentContext.wallet.generateNonce()),
                user_pin_required: (_b = preAuthorizedCodeFlowConfig.userPinRequired) !== null && _b !== void 0 ? _b : false,
            },
            authorization_code: authorizationCodeFlowConfig && {
                issuer_state: (_c = authorizationCodeFlowConfig.issuerState) !== null && _c !== void 0 ? _c : (await agentContext.wallet.generateNonce()),
            },
        };
        return grants;
    }
    findOfferedCredentialsMatchingRequest(credentialOffer, credentialRequest, credentialsSupported) {
        const offeredCredentials = (0, shared_1.getOfferedCredentials)(credentialOffer.credentials, credentialsSupported);
        return offeredCredentials.filter((offeredCredential) => {
            if (offeredCredential.format !== credentialRequest.format)
                return false;
            if (credentialRequest.format === shared_1.OpenId4VciCredentialFormatProfile.JwtVcJson &&
                offeredCredential.format === credentialRequest.format) {
                return (0, core_1.equalsIgnoreOrder)(offeredCredential.types, credentialRequest.types);
            }
            else if (credentialRequest.format === shared_1.OpenId4VciCredentialFormatProfile.JwtVcJsonLd &&
                offeredCredential.format === credentialRequest.format) {
                return (0, core_1.equalsIgnoreOrder)(offeredCredential.types, credentialRequest.credential_definition.types);
            }
            else if (credentialRequest.format === shared_1.OpenId4VciCredentialFormatProfile.LdpVc &&
                offeredCredential.format === credentialRequest.format) {
                return (0, core_1.equalsIgnoreOrder)(offeredCredential.types, credentialRequest.credential_definition.types);
            }
            else if (credentialRequest.format === shared_1.OpenId4VciCredentialFormatProfile.SdJwtVc &&
                offeredCredential.format === credentialRequest.format) {
                return offeredCredential.vct === credentialRequest.vct;
            }
            return false;
        });
    }
    async getHolderBindingFromRequest(credentialRequest) {
        var _a;
        if (!((_a = credentialRequest.proof) === null || _a === void 0 ? void 0 : _a.jwt))
            throw new core_1.CredoError('Received a credential request without a proof');
        const jwt = core_1.Jwt.fromSerializedJwt(credentialRequest.proof.jwt);
        if (jwt.header.kid) {
            if (!jwt.header.kid.startsWith('did:')) {
                throw new core_1.CredoError("Only did is supported for 'kid' identifier");
            }
            else if (!jwt.header.kid.includes('#')) {
                throw new core_1.CredoError(`kid containing did MUST point to a specific key within the did document: ${jwt.header.kid}`);
            }
            return {
                method: 'did',
                didUrl: jwt.header.kid,
            };
        }
        else if (jwt.header.jwk) {
            return {
                method: 'jwk',
                jwk: (0, core_1.getJwkFromJson)(jwt.header.jwk),
            };
        }
        else {
            throw new core_1.CredoError('Either kid or jwk must be present in credential request proof header');
        }
    }
};
OpenId4VcIssuerService = __decorate([
    (0, core_1.injectable)(),
    __metadata("design:paramtypes", [core_1.W3cCredentialService,
        core_1.JwsService,
        OpenId4VcIssuerModuleConfig_1.OpenId4VcIssuerModuleConfig,
        repository_1.OpenId4VcIssuerRepository])
], OpenId4VcIssuerService);
exports.OpenId4VcIssuerService = OpenId4VcIssuerService;
//# sourceMappingURL=OpenId4VcIssuerService.js.map